@@ -43,21 +43,33 @@ pub enum Card {
     CallNative(Box<CallNode>),
     IfTrue(LaneNode),
     IfFalse(LaneNode),
-    IfElse { then: LaneNode, r#else: LaneNode },
+    IfElse {
+        then: LaneNode,
+        r#else: LaneNode,
+    },
     Jump(LaneNode),
     SetGlobalVar(VarNode),
     SetVar(VarNode),
     ReadVar(VarNode),
     Repeat(LaneNode),
     While(LaneNode),
-    ForEach { variable: VarNode, lane: LaneNode },
+    ForEach {
+        variable: VarNode,
+        lane: LaneNode,
+    },
+    /// Single card that decomposes into multiple cards
+    // TODO: move into struct and store a Box<CompositeCard> to reduce Card size?
+    CompositeCard {
+        name: String,
+        cards: Vec<Card>,
+    },
+    /// Does nothing
+    Noop,
 }
 
 impl Card {
-    pub fn name(&self) -> &'static str {
+    pub fn name(&self) -> &str {
         match self {
-            // Card::GetByKey => "GetByKey",
-            // Card::SetByKey => "SetByKey",
             Card::SetVar(_) => "SetLocalVar",
             Card::Pass => "Pass",
             Card::Add => "Add",
@@ -95,6 +107,8 @@ impl Card {
             Card::GetProperty => "GetProperty",
             Card::SetProperty => "SetProperty",
             Card::ForEach { .. } => "ForEach",
+            Card::CompositeCard { name, .. } => name.as_str(),
+            Card::Noop => "No-op",
         }
     }
 
@@ -107,7 +121,9 @@ impl Card {
             | Card::SetVar(_)
             | Card::While(_)
             | Card::Repeat(_)
-            | Card::ForEach { .. } => None,
+            | Card::ForEach { .. }
+            | Card::CompositeCard { .. }
+            | Card::Noop => None,
 
             Card::GetProperty => Some(Instruction::GetProperty),
             Card::SetProperty => Some(Instruction::SetProperty),
@@ -212,35 +228,37 @@ pub struct SubProgramNode(pub InputString);
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct StringNode(pub String);
 
-#[derive(Debug, Clone, Default, Copy)]
+#[derive(Debug, Clone, Default)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
-pub struct VarNode(pub VarName);
+pub struct VarNode(pub Box<VarName>);
 
 impl VarNode {
     /// panics if the string is too long
     pub fn from_str_unchecked(s: &str) -> Self {
-        Self(VarName::from(s).expect("Failed to parse variable name"))
+        Self(Box::new(
+            VarName::from(s).expect("Failed to parse variable name"),
+        ))
     }
 }
 
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
-pub enum LaneNode {
-    LaneName(String),
-    LaneId(usize),
-}
+pub struct LaneNode(pub String);
 
 impl Display for LaneNode {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        match self {
-            LaneNode::LaneName(n) => write!(f, "{}", n),
-            LaneNode::LaneId(n) => write!(f, "#{}", n),
-        }
+        write!(f, "{}", self.0)
     }
 }
 
-impl Default for LaneNode {
-    fn default() -> Self {
-        Self::LaneId(0)
+impl From<LaneNode> for Handle {
+    fn from(ln: LaneNode) -> Self {
+        Self::from(&ln)
+    }
+}
+
+impl<'a> From<&'a LaneNode> for Handle {
+    fn from(ln: &'a LaneNode) -> Self {
+        Handle::from_str(ln.0.as_str()).unwrap()
     }
 }
