@@ -7,16 +7,14 @@ def test_compile_and_run():
     """
 
     PROGRAM_YAML = """
-module:
-    lanes:
-        main: 
-            name: main
-            cards:
-                - ty: ScalarInt
-                  val: 5
-                - ty: ScalarInt
-                  val: 5
-                - ty: Add
+lanes:
+    main: 
+        cards:
+            - ty: ScalarInt
+              val: 5
+            - ty: ScalarInt
+              val: 5
+            - ty: Add
 """
 
     program = caoc.CompilationUnit.from_yaml(PROGRAM_YAML)
@@ -30,3 +28,23 @@ module:
 def test_get_version():
     v = caoc.native_version()
     assert isinstance(v, str)
+
+
+def test_json():
+    PROGRAM_JSON = """
+    {
+        "lanes": {
+            "main": {
+                "cards": [
+                    { "ty": "Noop" }
+                ]
+            }
+        }
+    }
+    """
+    program = caoc.CompilationUnit.from_json(PROGRAM_JSON)
+    options = caoc.CompilationOptions()
+
+    program = caoc.compile(program, options)
+
+    caoc.run(program)
